#!/usr/bin/env python3
import importlib.util
import logging
import logging.config
import os
import subprocess
from logging import Logger
from pathlib import Path
from typing import Any, Dict

import uvicorn  # type: ignore


def set_conf_path(module_stem: str) -> str:
    """Set the path to a configuration file."""
    conf_var = str(os.getenv(f"{module_stem.upper()}_CONF"))
    if Path(conf_var).is_file():
        conf_path = conf_var
    elif Path(f"/app/inboard/{module_stem}_conf.py").is_file():
        conf_path = f"/app/inboard/{module_stem}_conf.py"
    else:
        conf_path = f"/{module_stem}_conf.py"
    os.environ[f"{module_stem.upper()}_CONF"] = conf_path
    return conf_path


def configure_logging(logger: Logger = logging.getLogger()) -> Dict[str, Any]:
    """Configure Python logging based on a path to a logging module or file."""
    try:
        logging_conf_path = str(os.getenv("LOGGING_CONF", "inboard.logging_conf"))
        spec = importlib.util.find_spec(logging_conf_path)
        logging_conf_module = importlib.util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(logging_conf_module)  # type: ignore
        if hasattr(logging_conf_module, "LOGGING_CONFIG"):
            logging_conf_dict = getattr(logging_conf_module, "LOGGING_CONFIG")
        else:
            raise AttributeError(f"No LOGGING_CONFIG in {logging_conf_module}.")
        if isinstance(logging_conf_dict, dict):
            logging.config.dictConfig(logging_conf_dict)
            message = f"Logging dict config loaded from {logging_conf_path}."
            logger.debug(message)
            return logging_conf_dict
        else:
            raise TypeError("LOGGING_CONFIG is not a dictionary instance.")
    except Exception as e:
        message = f"Error when setting logging module: {e}."
        logger.debug(message)
        raise


def set_app_module(logger: Logger = logging.getLogger()) -> str:
    """Set the name of the Python module with the app instance to run."""
    try:
        app_module = str(os.getenv("APP_MODULE"))
        importlib.util.find_spec(app_module)
        logger.debug(f"App module set to {app_module}.")
        return app_module
    except Exception as e:
        message = f"Error when setting app module: {e}."
        logger.debug(message)
        raise


def run_pre_start_script(logger: Logger = logging.getLogger()) -> str:
    """Run a pre-start script at the provided path."""
    try:
        logger.debug("Checking for pre-start script.")
        pre_start_path = os.getenv("PRE_START_PATH", "/app/inboard/app/prestart.py")
        if Path(pre_start_path).is_file():
            process = "python" if Path(pre_start_path).suffix == ".py" else "sh"
            run_message = f"Running pre-start script with {process} {pre_start_path}."
            logger.debug(run_message)
            subprocess.run([process, pre_start_path])
            message = f"Ran pre-start script with {process} {pre_start_path}."
        else:
            message = "No pre-start script found."
    except Exception as e:
        message = f"Error from pre-start script: {e}."
    finally:
        logger.debug(message)
        return message


def start_server(
    app_module: str = str(os.getenv("APP_MODULE", "inboard.app.base.main:app")),
    gunicorn_conf: str = str(
        os.getenv("GUNICORN_CONF", "/app/inboard/gunicorn_conf.py")
    ),
    logger: Logger = logging.getLogger(),
    logging_conf_dict: Dict[str, Any] = None,
    process_manager: str = str(os.getenv("PROCESS_MANAGER", "gunicorn")),
    with_reload: bool = bool(os.getenv("WITH_RELOAD", False)),
    worker_class: str = str(os.getenv("WORKER_CLASS", "uvicorn.workers.UvicornWorker")),
) -> None:
    """Start the Uvicorn or Gunicorn server."""
    try:
        if process_manager == "gunicorn":
            logger.debug("Running Uvicorn with Gunicorn.")
            gunicorn_conf_path = Path(gunicorn_conf)
            subprocess.run(
                ["gunicorn", "-k", worker_class, "-c", gunicorn_conf_path, app_module]
            )
        elif process_manager == "uvicorn":
            logger.debug("Running Uvicorn without Gunicorn.")
            uvicorn.run(
                app_module,
                host=os.getenv("HOST", "0.0.0.0"),
                port=int(os.getenv("PORT", "80")),
                log_config=logging_conf_dict,
                log_level=os.getenv("LOG_LEVEL", "info"),
                reload=with_reload,
            )
        else:
            raise NameError("Process manager needs to be either uvicorn or gunicorn.")
    except Exception as e:
        logger.debug(f"Error when starting server with start script: {e}")
        raise


if __name__ == "__main__":
    logger = logging.getLogger()
    logging_conf_dict = configure_logging(logger=logger)
    gunicorn_conf_path = set_conf_path("gunicorn")
    app_module = set_app_module(logger=logger)
    run_pre_start_script(logger=logger)
    start_server(
        app_module=app_module,
        gunicorn_conf=gunicorn_conf_path,
        logger=logger,
        logging_conf_dict=logging_conf_dict,
    )



import os
import sys
from typing import Dict, Union

LOG_COLORS = bool(os.getenv("LOG_COLORS", sys.stdout.isatty()))
LOG_FORMAT = str(os.getenv("LOG_FORMAT", "simple"))
LOG_LEVEL = str(os.getenv("LOG_LEVEL", "info")).upper()
LOGGING_CONFIG: Dict[str, Union[Dict, bool, int, str]] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {
            "class": "logging.Formatter",
            "format": "%(levelname)-10s %(message)s",
        },
        "verbose": {
            "class": "logging.Formatter",
            "format": "%(asctime)-30s %(process)-10d %(name)-15s %(module)-15s %(levelname)-10s %(message)s",  # noqa: E501
            "datefmt": "%Y-%m-%d %H:%M:%S %z",
        },
        "gunicorn": {
            "class": "logging.Formatter",
            "format": "%(asctime)s [%(process)d] [%(levelname)s] %(message)s",
            "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
        },
        "uvicorn": {
            "()": "uvicorn.logging.DefaultFormatter",
            "format": "%(levelprefix)s %(message)s",
            "use_colors": LOG_COLORS,
        },
    },
    "handlers": {
        "default": {
            "class": "logging.StreamHandler",
            "formatter": LOG_FORMAT,
            "level": LOG_LEVEL,
            "stream": "ext://sys.stdout",
        }
    },
    "root": {"handlers": ["default"], "level": LOG_LEVEL},
    "loggers": {
        "fastapi": {"propagate": True},
        "gunicorn.access": {"handlers": ["default"], "propagate": True},
        "gunicorn.error": {"propagate": True},
        "uvicorn": {"propagate": True},
        "uvicorn.access": {"propagate": True},
        "uvicorn.asgi": {"propagate": True},
        "uvicorn.error": {"propagate": True},
    },
}



import json
import multiprocessing
import os
from pathlib import Path

from inboard.start import configure_logging

workers_per_core_str = os.getenv("WORKERS_PER_CORE", "1")
max_workers_str = os.getenv("MAX_WORKERS")
use_max_workers = None
if max_workers_str:
    use_max_workers = int(max_workers_str)
web_concurrency_str = os.getenv("WEB_CONCURRENCY", None)

host = os.getenv("HOST", "0.0.0.0")
port = os.getenv("PORT", "80")
bind_env = os.getenv("BIND", None)
use_loglevel = os.getenv("LOG_LEVEL", "info")
use_bind = bind_env if bind_env else f"{host}:{port}"
cores = multiprocessing.cpu_count()
workers_per_core = float(workers_per_core_str)
default_web_concurrency = workers_per_core * cores
if web_concurrency_str:
    web_concurrency = int(web_concurrency_str)
    assert web_concurrency > 0
else:
    web_concurrency = max(int(default_web_concurrency), 2)
    if use_max_workers:
        web_concurrency = min(web_concurrency, use_max_workers)
accesslog_var = os.getenv("ACCESS_LOG", "-")
use_accesslog = accesslog_var or None
errorlog_var = os.getenv("ERROR_LOG", "-")
use_errorlog = errorlog_var or None
graceful_timeout_str = os.getenv("GRACEFUL_TIMEOUT", "120")
timeout_str = os.getenv("TIMEOUT", "120")
keepalive_str = os.getenv("KEEP_ALIVE", "5")

# Gunicorn config variables
try:
    logconfig_dict = configure_logging()
except Exception as e:
    if use_loglevel == "debug":
        msg = "Error loading logging config with Gunicorn:"
        print(f"[{Path(__file__).stem}] {msg} {e}")
loglevel = use_loglevel
workers = web_concurrency
bind = use_bind
errorlog = use_errorlog
worker_tmp_dir = "/dev/shm"
accesslog = use_accesslog
graceful_timeout = int(graceful_timeout_str)
timeout = int(timeout_str)
keepalive = int(keepalive_str)

log_data = {
    # General
    "host": host,
    "port": port,
    "use_max_workers": use_max_workers,
    "workers_per_core": workers_per_core,
    # Gunicorn
    "loglevel": loglevel,
    "workers": workers,
    "bind": bind,
    "graceful_timeout": graceful_timeout,
    "timeout": timeout,
    "keepalive": keepalive,
    "errorlog": errorlog,
    "accesslog": accesslog,
}
if loglevel == "debug":
    print(f"[{Path(__file__).stem}] Custom configuration:", json.dumps(log_data))



"""
inboard
---
"""
from importlib.metadata import version


def package_version() -> str:
    """Calculate version number based on pyproject.toml"""
    try:
        return version(__package__)
    except Exception:
        return "Package not found."


__version__ = package_version()
